package main

import (
	"flag"
	"fmt"
	"github.com/go-kratos/kratos/contrib/registry/consul/v2"
	"github.com/go-kratos/kratos/v2"
	"github.com/go-kratos/kratos/v2/middleware/tracing"
	"github.com/go-kratos/kratos/v2/transport/grpc"
	"go.opentelemetry.io/otel"
	"go.opentelemetry.io/otel/exporters/jaeger"
	"go.opentelemetry.io/otel/sdk/resource"
	tracesdk "go.opentelemetry.io/otel/sdk/trace"
	semconv "go.opentelemetry.io/otel/semconv/v1.10.0"
	"go.uber.org/zap"
	"hilo/app/service/user/internal/conf"
	"hilo/common/aws"
	"hilo/common/logz"
	"hilo/common/tencent"
	"hilo/common/util"
	"math/rand"
	"time"

	"github.com/go-kratos/kratos/v2/log"
)

var (
	id = util.GenServiceId()
)

func main() {
	flag.Parse()
	conf.LoadConfig()
	rand.Seed(time.Now().UnixNano())

	err := initTracer(conf.Conf.Jaeger.Url)
	if err != nil {
		panic(err)
	}

	l := logz.NewLogger(conf.Conf.Log, zap.AddCaller(), zap.AddCallerSkip(3))
	defer l.Flush()

	logger := log.With(l,
		"service.id", id,
		"service.name", conf.Conf.Server.Name,
		"service.version", conf.Conf.Server.Version,
		"trace.id", tracing.TraceID(),
		"span.id", tracing.SpanID(),
	)
	log.SetLogger(logger)

	aws.SetAws(aws.NewAWS(conf.Conf.Aws))
	tencent.SetTencentYun(tencent.NewTencentYun(conf.Conf.TencentYun))

	app, cleanup, err := wireApp(
		conf.Conf.Server, conf.Conf.Client, conf.Conf.Data, conf.Conf.Consul,
		logger,
	)
	if err != nil {
		panic(err)
	}
	defer cleanup()

	// start and wait for stop signal
	if err := app.Run(); err != nil {
		panic(err)
	}
}

func newApp(logger log.Logger, gs *grpc.Server, rr *consul.Registry) *kratos.App {
	return kratos.New(
		kratos.ID(id),
		kratos.Name(conf.Conf.Server.Name),
		kratos.Version(conf.Conf.Server.Version),
		kratos.Metadata(map[string]string{}),
		kratos.Logger(logger),
		kratos.Server(
			gs,
		),
		kratos.Registrar(rr),
	)
}

// 设置全局trace
func initTracer(url string) error {
	// 创建 Jaeger exporter
	exp, err := jaeger.New(jaeger.WithCollectorEndpoint(jaeger.WithEndpoint(url)))
	if err != nil {
		return err
	}
	tp := tracesdk.NewTracerProvider(
		// 将基于父span的采样率设置为100%
		tracesdk.WithSampler(tracesdk.ParentBased(tracesdk.TraceIDRatioBased(1.0))),
		// 始终确保再生成中批量处理
		tracesdk.WithBatcher(exp),
		// 在资源中记录有关此应用程序的信息
		tracesdk.WithResource(resource.NewSchemaless(
			semconv.ServiceNameKey.String(fmt.Sprintf("kratos-trace-%s", conf.Conf.Server.Name)),
			//attribute.String("exporter", "jaeger"),
			//attribute.Float64("float", 312.23),
		)),
	)

	otel.SetTracerProvider(tp)
	return nil
}
