package service

import (
	"context"
	"crypto/rsa"
	"encoding/base64"
	"encoding/json"
	"github.com/golang-jwt/jwt/v4"
	"github.com/spf13/cast"
	"github.com/tidwall/gjson"
	"google.golang.org/grpc/peer"
	"hilo/api/service/country"
	"hilo/api/service/mgr"
	"hilo/api/service/user"
	"hilo/app/service/user/internal/conf"
	"hilo/app/service/user/internal/model"
	"hilo/app/service/user/internal/pkg"
	"hilo/common/aws"
	"hilo/common/errm"
	"hilo/common/httpclient"
	"hilo/common/mysql"
	"hilo/common/tencent"
	"hilo/common/util"
	"math/big"
	"net"
	"strings"
	"time"
)

const (
	PublicKeyReqUrl     = "https://appleid.apple.com/auth/keys"
	AppleUrl            = "https://appleid.apple.com"
	ApplicationClientId = "com.qiahao.nextvideo"
)

type JwtClaims struct {
	jwt.StandardClaims
}

type JwtHeader struct {
	Kid string `json:"kid"`
	Alg string `json:"alg"`
}

type JwtKeys struct {
	Kty string `json:"kty"`
	Kid string `json:"kid"`
	Use string `json:"use"`
	Alg string `json:"alg"`
	N   string `json:"n"`
	E   string `json:"e"`
}

func (s *UserService) Login(ctx context.Context, req *user.LoginReq) (*user.LoginRep, error) {
	var imei string
	if imei != "" {
		res, err := s.dao.MgrClient.CheckMgrImeiExist(ctx, &mgr.CheckMgrImeiExistReq{Imei: imei})
		if err != nil {
			return nil, errm.CommonError(err.Error())
		}
		if res.Ok == false {
			return nil, UserFreeze
		}
	}

	var ip string
	if pr, ok := peer.FromContext(ctx); ok {
		if tcpAddr, ok := pr.Addr.(*net.TCPAddr); ok {
			ip = tcpAddr.IP.String()
		} else {
			ip = pr.Addr.String()
		}
	}
	//if ip != "" {
	//	resp, err := s.dao.MgrClient.CheckMgrIpExist(ctx, &mgr.CheckMgrIpExistReq{Ip: ip})
	//	if err != nil {
	//		return nil, errm.CommonError(err.Error())
	//	}
	//	if resp.Ok == false {
	//		s.log.WithContext(ctx).Error(UserFreeze)
	//		return nil, UserFreeze
	//	}
	//}

	var (
		userInfo *model.User
	)

	flag, token, userInfo, err := s.logoutCheckOauth(req.ThirdPartyId, int(req.ThirdPartyType))
	if err != nil {
		return nil, err
	}

	if conf.AppIsRelease() {
		if ok := s.verifyAuthToken(req.ThirdPartyId, req.ThirdPartyType, req.ThirdPartyToken, token); !ok {
			if req.ThirdPartyType == user.EThirdPartyType_Phone {
				return nil, PhoneTokenFail
			} else {
				return nil, AuthAccessTokenFail
			}
		}
	}

	if !flag {
		//不是空图片，上传到AWS
		if req.Avatar != "" {
			image, err := aws.Global.UploadImage(req.Avatar)
			if err != nil {
				return nil, err
			} else {
				req.Avatar = image
			}
		}

		// 注册
		userInfo, err = s.register(ctx, imei, ip, req)
		if err != nil {
			return nil, err
		}
		//userId = u.Id

	} else {
		// 登陆成功
		// 更新平台信息
		userInfo.DeviceType = pkg.GetDeviceType(ctx)
		_, err = s.dao.UpdateUser(userInfo.Id, userInfo)
		if err != nil {
			return nil, err
		}
	}

	// 可以异步执行, 错误不处理
	err2 := s.dao.AddUserIp(userInfo.Id, ip)
	if err2 != nil {
		s.log.Errorf("add user ip err : %s", err2)
	}

	mgrStatus, eTime, err := s.SyncMgrUserStatus(ctx, userInfo.Id)

	if err != nil {
		return nil, err
	}
	if mgrStatus == mgr.UserStatus_Freeze {
		return nil, UserFreeze.WithMetadata(map[string]string{"timestamp": cast.ToString(eTime)})
	}

	txSig, err := tencent.Global.GetUserSig(userInfo.ExternalId)
	if err != nil {
		return nil, err
	}

	// 兼容于ios就版本，不然ios就退登 // 融云 TODO
	//if strings.Index(c.GetHeader("User-Agent"), "Hilo/1.4.1 (iPhone") >= 0 {
	//	ryToken, _ = rongyun.Auth(string(user.ExternalId), string(paramUserLogin.Nick), string(paramUserLogin.Avatar))
	//}

	jwtToken, err := pkg.GenerateJwtToken(userInfo.Id, userInfo.ExternalId)
	if err != nil {
		return nil, err
	}

	detail, err := s.GetUserDetail(ctx, userInfo)
	if err != nil {
		return nil, err
	}

	res := &user.LoginRep{
		Token:           jwtToken,
		RongyunToken:    txSig,
		TencentyunToken: txSig,
		ThirdPartyType:  req.ThirdPartyType,
		User: &user.CvUserDetail{
			ILikeCount:        int32(detail.ILikeCount),
			LikeCount:         0,
			VisitCount:        0,
			IsPush:            0,
			DiamondNum:        0,
			IsLike:            false,
			IsSession:         false,
			IsLikeMe:          false,
			HeartValue:        0,
			HeartValueMax:     0,
			MeetDays:          0,
			WealthUserGrade:   0,
			CharmUserGrade:    0,
			ActivityUserGrade: 0,
			CurrentRoom:       "",
			MyGroupPower:      0,
			MyGroupPowerName:  "",
			Info: &user.UserBase{
				Id:                            0,
				Avatar:                        "",
				DefaultAvatar:                 false,
				ExternalId:                    "",
				Nick:                          "",
				Description:                   "",
				ESex:                          0,
				Country:                       "",
				CountryIcon:                   "",
				Code:                          "",
				IsPrettyCode:                  false,
				IsLogout:                      false,
				Birthday:                      0,
				IsShowAge:                     false,
				IsTradeUnion:                  false,
				IsAgentMgr:                    false,
				IsTradeUnionMatchNotification: false,
				IsVip:                         false,
				IsOfficialStaff:               false,
				VipExpireTime:                 0,
				Medals:                        nil,
				MedalInfo:                     nil,
				Headwear:                      nil,
				Ride:                          nil,
				Noble:                         nil,
			},
		},
	}

	return res, nil
}

func (s *UserService) SyncMgrUserStatus(ctx context.Context, userId int64) (mgr.UserStatus, int64, error) {
	res, err := s.dao.MgrClient.GetMgrUserStatus(ctx, &mgr.GetMgrUserStatusReq{UserId: userId})
	if err != nil {
		return mgr.UserStatus_None, 0, err
	}

	if res.Status == mgr.UserStatus_Normal {
		if res.EndTime == nil {
			return mgr.UserStatus_Freeze, 0, nil
		}

		now := time.Now()
		et := res.EndTime.AsTime()
		if now.After(et) {
			_, err2 := s.dao.UpdateUser(userId, &model.User{Status: model.UserNormalStatus})
			if err2 != nil {
				s.log.Errorf("UpdateUser status err: %s", err2)
			}
			return mgr.UserStatus_Normal, 0, nil
		} else {
			return mgr.UserStatus_Freeze, et.Unix() - now.Unix(), nil
		}
	}

	return mgr.UserStatus_Normal, 0, nil
}

func (s *UserService) register(ctx context.Context, imei string, ip string, req *user.LoginReq) (*model.User, error) {
	if conf.AppIsRelease() {
		if imei != "" {
			if c, err := s.dao.CountImei(imei); err != nil {
				return nil, errm.DbError(err.Error())
			} else if int(c) > conf.Conf.Auth.ImeiTotalLimit {
				return nil, RegisterFail
			}

			if c, err := s.dao.CountImeiThird(imei, int(req.ThirdPartyType)); err != nil {
				return nil, errm.DbError(err.Error())
			} else if int(c) > conf.Conf.Auth.ImeiOauthLimit {
				return nil, RegisterFail
			}
		}

		if ip != "" {
			beginTime := time.Now().Add(-1 * time.Duration(conf.Conf.Auth.AccountIpDurationLimit) * time.Second)
			if c, err := s.dao.CountIp(ip, beginTime); err != nil {
				return nil, errm.DbError(err.Error())
			} else if int(c) > conf.Conf.Auth.AccountIpLimit {
				return nil, RegisterFail
			}
		}
	}

	//userOauth := &model.UserOauth{
	//	ThirdPartyId:    req.ThirdPartyId,
	//	ThirdPartyType:  int(req.ThirdPartyType),
	//	ThirdPartyToken: req.ThirdPartyToken,
	//}
	if req.Avatar != "" {
		req.Avatar = pkg.MakeFullUrl(req.Avatar)
	}

	countryIcon := ""
	if req.Country != "" {
		res, err := s.dao.CountryClient.GetCountryByShortName(ctx, &country.GetCountryByShortNameReq{
			ShortName: req.Country,
		})
		if err != nil {
			return nil, err
		}
		countryIcon = res.Data.Icon
		req.Country = res.Data.Name
	}
	u := &model.User{
		ExternalId:    util.GetRandStrByUUID(),
		Avatar:        req.Avatar,
		DefaultAvatar: 1,
		Nick:          req.Nick,
		Sex:           int(req.Sex),
		Birthday:      0,
		Country:       req.Country,
		CountryIcon:   countryIcon,
		Description:   "",
		IsPush:        int(mysql.CLOSE),
		IsShowAge:     int(mysql.OPEN),
		DeviceType:    pkg.GetDeviceType(ctx),
	}

	cu, err := s.dao.CreateUser(u)
	if err != nil {
		return nil, errm.DbError(err.Error())
	}

	err = tencent.Global.UserRegister(ctx, cu.ExternalId, cu.Nick, cu.Avatar)
	if err != nil {
		s.log.WithContext(ctx).Errorf("tencent register err : %s", err)
		return nil, err
	}
	// TODO: 新增金币等事件发布, 服务间异步通信可以考虑broker

	return u, nil
}

func (s *UserService) logoutCheckOauth(thirdPartyId string, thirdPartyType int) (
	flag bool, phoneToken string, u *model.User, err error,
) {
	userOauth, err2 := s.dao.GetUserOauth(thirdPartyId, thirdPartyType)
	if err2 != nil {
		err = err2
		return
	}

	if userOauth == nil {
		return
	}

	u, err2 = s.dao.GetUser(userOauth.UserId)
	if err2 != nil {
		err = err2
		return
	}

	if u.LogoutTime > 0 && (time.Now().Unix() > u.LogoutTime) {
		//userOauth.ThirdPartyId = userOauth.ThirdPartyId + "&" + pkg.GetRandStrByUUID()
		userOauth.ThirdPartyId = ""
		_, err2 := s.dao.UpdateUserOauth(u.Id, userOauth)
		if err2 != nil {
			err = err2
			return
		}
	} else {
		phoneToken = userOauth.ThirdPartyToken
		flag = true
	}

	return
}

func (s *UserService) verifyAuthToken(thirdPartyId string, thirdPartyType user.EThirdPartyType, thirdPartyToken string, phoneToken string) bool {
	switch thirdPartyType {
	case user.EThirdPartyType_Google:
		return s.verifyGoogleToken(thirdPartyToken)
	case user.EThirdPartyType_Facebook:
		return s.verifyFaceBookToken(thirdPartyId, thirdPartyToken)
	case user.EThirdPartyType_Apple:
		return s.verifyAppleIdentityToken(thirdPartyToken, thirdPartyId)
	case user.EThirdPartyType_WeChat:
		return s.verifyWechatToken(thirdPartyId, thirdPartyToken)
	case user.EThirdPartyType_Phone:
		/*if (thirdPartyId == "12610000999" && thirdPartyToken == "Hilo2020") || (thirdPartyId == "12610000001" && thirdPartyToken == "Hilo@05#A") {
			return true
		}*/
		//同时拦截phone的注册，不允许phone的注册，想通过phone
		if thirdPartyToken != "" && phoneToken != "" && thirdPartyToken == phoneToken {
			return true
		}
	}
	return false
}

/**
* 校验wechat
**/
func (s *UserService) verifyWechatToken(thirdPartyId string, thirdPartyToken string) bool {
	url := "https://api.weixin.qq.com/sns/auth?access_token=" + thirdPartyToken + "&openid=" + thirdPartyId
	resString, err := httpclient.GetRemoteContent(url)

	if err != nil {
		s.log.Error(err)

		return false
	}
	errcode := gjson.Get(resString, "errcode").String()
	if errcode != "0" {
		s.log.Error("wechat 校验失败，结果:", errcode, " opneId:", thirdPartyId, " token:", thirdPartyToken)
		return false
	}
	return true
}

func (s *UserService) verifyFaceBookToken(thirdPartyId string, thirdPartyToken string) bool {
	url := "https://graph.facebook.com/" + thirdPartyId + "?fields=id,name,gender,picture&access_token=" + thirdPartyToken
	resString, err := httpclient.GetRemoteContent(url)
	if err != nil {
		s.log.Error(err)
		return false
	}

	infoID := gjson.Get(resString, "id").String()
	if len(infoID) > 0 {
		return true
	}
	return false
}

func (s *UserService) verifyGoogleToken(thirdPartyToken string) bool {
	tokenEncode, explodeBool := explodeToken(thirdPartyToken)
	if !explodeBool {
		s.log.Errorf("google token 校验失败：%v", thirdPartyToken)
		return false
	}

	tokenDecode, err := base64.RawURLEncoding.DecodeString(tokenEncode)

	if err != nil {
		s.log.Error(err)
		return false
	}

	tokenDecodeJson := string(tokenDecode)

	androidTokenRes := gjson.GetMany(tokenDecodeJson, "iss", "azp", "aud", "sub")

	if androidTokenRes[0].String() == "https://accounts.google.com" &&
		len(androidTokenRes[1].String()) > 0 &&
		len(androidTokenRes[2].String()) > 0 && len(androidTokenRes[3].String()) > 0 {
		return true
	}

	return true
}

func explodeToken(thirdPartyToken string) (string, bool) {
	tokens := strings.Split(thirdPartyToken, ".")

	if len(tokens) != 3 {
		return "", false
	}

	return tokens[1], true
}

// 认证客户端传递过来的token是否有效
func (s *UserService) verifyAppleIdentityToken(cliToken string, cliUserID string) bool {
	// 数据由 头部、载荷、签名 三部分组成
	cliTokenArr := strings.Split(cliToken, ".")
	if len(cliTokenArr) < 3 {
		//syslog.Logger().Errorln("cliToken Split err ! cliToken = ", cliToken)
		s.log.Error("cliToken Split err")
		return false
	}

	// 解析cliToken的header获取kid
	cliHeader, err := jwt.DecodeSegment(cliTokenArr[0])
	if err != nil {
		//syslog.Logger().Errorln(err.Error())
		s.log.Errorf("VerifyIdentityToken err :%v", err)
		return false
	}

	var jHeader JwtHeader
	err = json.Unmarshal(cliHeader, &jHeader)
	if err != nil {
		//syslog.Logger().Errorln(err.Error())
		s.log.Errorf("VerifyIdentityToken err :%v", err)
		return false
	}

	// 效验pubKey 及 token
	token, err := jwt.ParseWithClaims(cliToken, &JwtClaims{}, func(token *jwt.Token) (interface{}, error) {
		return GetRSAPublicKey(jHeader.Kid), nil
	})

	if err != nil {
		//syslog.Logger().Errorln(err.Error())
		s.log.Errorf("VerifyIdentityToken err :%v", err)
		return false
	}

	// 信息验证
	if claims, ok := token.Claims.(*JwtClaims); ok && token.Valid {
		if claims.Issuer != AppleUrl || claims.Audience != ApplicationClientId || claims.Subject != cliUserID {
			//syslog.Logger().Errorln("verify token info fail, info is not match")
			s.log.Errorf("verify token info fail, info is not match")
			return false
		}
		// here is verify ok !
	} else {
		s.log.Errorf("token claims parse fail")
		return false
	}

	return true
}

// GetRSAPublicKey 向苹果服务器获取解密signature所需要用的publicKey
func GetRSAPublicKey(kid string) *rsa.PublicKey {
	response, err := httpclient.GetRemoteBody(PublicKeyReqUrl)
	if err != nil {
		//syslog.Logger().Errorln(err.Error())
		return nil
	}

	var jKeys map[string][]JwtKeys
	err = json.Unmarshal(response, &jKeys)
	if err != nil {
		//syslog.Logger().Errorln(err.Error())
		return nil
	}

	// 获取验证所需的公钥
	var pubKey rsa.PublicKey
	// 通过cliHeader的kid比对获取n和e值 构造公钥
	for _, data := range jKeys {
		for _, val := range data {
			if val.Kid == kid {
				nBin, _ := base64.RawURLEncoding.DecodeString(val.N)
				nData := new(big.Int).SetBytes(nBin)

				eBin, _ := base64.RawURLEncoding.DecodeString(val.E)
				eData := new(big.Int).SetBytes(eBin)

				pubKey.N = nData
				pubKey.E = int(eData.Uint64())
				break
			}
		}
	}

	if pubKey.E <= 0 {
		//syslog.Logger().Errorln("rsa.PublicKey get fail !")
		return nil
	}
	return &pubKey
}
